// BluetoothPeer.java

/*
This software is part of the Aplu Bluetooth Library.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
 */

package ch.aplu.bluetooth;

import javax.bluetooth.*;
import java.io.*;

/**
 * Class that implements a Bluetooth peer-to-peer communication based on
 * the client-server model. When started, an existing server is searched
 * using a specified Bluetooth service. If found, the connection is establish
 * in a client mode. Otherwise a server mode with the given service name is started.<br><br>
 *
 * Data is exchanged as integer blocks of arbitrary length.
 */
public class BluetoothPeer
{

  // --------------- Inner class MyBtListener --------------
  private class MyBtListener implements BtListener
  {
    // I am a server and a client connected successfully
    public void notifyConnection(RemoteDevice rd, InputStream is, OutputStream os)
    {
      if (isVerbose)
      {
        try
        {
          VerboseWriter.out.println("Peer: Connection established.");
          VerboseWriter.out.println("Peer: Client's name: " + rd.getFriendlyName(false));
          VerboseWriter.out.println("Peer: Client's address " + rd.getBluetoothAddress());
        }
        catch (IOException ex)
        {
        }
      }
      isConnected = true;
      if (listener != null)
        listener.notifyConnection(true);
      dis = new DataInputStream(is);
      dos = new DataOutputStream(os);

      int value = 0;
      int size = 0;
      int[] data = null;
      int index = -1;
      try
      {
        while (isConnected)
        {
          if (isVerbose)
            VerboseWriter.out.println("Peer: Calling readInt()");

          // Blocking, throws IOException
          // when client closes connection or BluetoothServer.cancel() is called
          value = dis.readInt();

          if (isVerbose)
            VerboseWriter.out.println("Peer: Returning from readInt(). Got: " + value);


          if (index == -1)
          {
            size = value;
            data = new int[size];
            index = 0;
          }
          else
          {
            data[index++] = value;
            if (index == size)
            {
              index = -1;
             if (listener != null)
               listener.receiveDataBlock(data);
            }
          }
        }
      }
      catch (IOException ex)
      {
        if (isVerbose)
          VerboseWriter.out.println("Peer: Connection lost");
        isConnected = false;
        if (listener != null)
          listener.notifyConnection(false);
        bs.cancel();
      }
    }
  }

  // I am a client and handle the connection
  private class ExecuteThread extends Thread
  {
    public void run()
    {
      dis = new DataInputStream(bc.getInputStream());
      dos = new DataOutputStream(bc.getOutputStream());

      boolean isRunning = true;
      boolean isError = false;

      int value = 0;
      int size = 0;
      int[] data = null;
      int index = -1;
      while (isRunning)
      {
        try
        {
          if (isVerbose)
            VerboseWriter.out.println("Peer: Call blocking readline()");
          value = dis.readInt(); // Blocking
        }
        catch (EOFException ex)
        {
          if (isVerbose)
            VerboseWriter.out.println("Peer: End of stream occured");
          isError = true;
        }
        catch (IOException ex)
        {
          if (isVerbose)
            VerboseWriter.out.println("Peer: Got IOException");
          isError = true;
        }
        if (isError)
        {
          bc.disconnect();
          isRunning = false;
        }
        else
        {
          if (isVerbose)
            VerboseWriter.out.println("Peer: Returned from blocking readline(). Got: " + value);
          if (index == -1)
          {
            size = value;
            data = new int[size];
            index = 0;
          }
          else
          {
            data[index++] = value;
            if (index == size)
            {
              index = -1;
              if (listener != null)
                listener.receiveDataBlock(data);
            }
          }
        }
      }
      if (listener != null)
        listener.notifyConnection(false);
    }
  }

  // --------------- End of inner classes ------------------

  private BtPeerListener listener = null;
  private String nodeName;
  private String serviceName;
  private BluetoothServer bs = null;
  private BluetoothClient bc = null;
  private DataInputStream dis;
  private DataOutputStream dos;
  private boolean isVerbose = false;
  private boolean isConnected = false;

  /**
   * Creates a BluetoothPeer instance that will connect to the given node using
   * the given service name. The given listener receives notifications about
   * connection status and incoming data blocks.<br><br>
   * @param nodeName the Bluetooth name of the partner node
   * @param serviceName the Bluetooth service name used for the communication channel
   * @param listener the BtPeerListener that receives notifications
   * @param isVerbose if true, debug information is sent using a VerboseWriter
   */
  public BluetoothPeer(String nodeName, String serviceName, BtPeerListener listener, boolean isVerbose)
  {
    this.nodeName = nodeName;
    this.serviceName = serviceName;
    this.listener = listener;
    this.isVerbose = isVerbose;
    if (isVerbose)
      VerboseWriter.out.println("Peer: Trying to connect as client to server " + nodeName +
        " using service " + serviceName);
    bc = new BluetoothClient(nodeName, serviceName);
    bc.setVerbose(isVerbose);
    if (bc.connect())
    {
      if (isVerbose)
        VerboseWriter.out.println("Peer: Connection as client successful");
      isConnected = true;
      listener.notifyConnection(true);
      new ExecuteThread().start();
    }
    else
    {
      if (isVerbose)
        VerboseWriter.out.println("Peer: Connection as client failed. Starting as server");
      bs = new BluetoothServer(serviceName, new MyBtListener(), isVerbose);
    }
  }


  /**
   * Send a block of data to the connected node. Returns immediately of not connected.
   * @param data the data block to send
   */
  public void sendDataBlock(int[] data)
  {
    if (isConnected)
    {
      try
      {
        int size = data.length;
        if (size == 0)
          return;
        dos.writeInt(size);
        dos.flush();
        if (isVerbose)
          VerboseWriter.out.print("Peer: Sending block\n" + size + ": ");
        for (int i = 0; i < size; i++)
        {
          dos.writeInt(data[i]);
          if (isVerbose)
            VerboseWriter.out.print(data[i] + ((i < size-1) ? "|" : "\n"));
          dos.flush();
        }
      }
      catch (IOException ex)
      {
        if (isVerbose)
          VerboseWriter.out.println("Peer: IOException in writeInt()");
      }
    }
  }

  /**
   * Returns the connection status.
   * @return true, if connected, otherwise false
   */
  public boolean isConnected()
  {
    return isConnected;
  }

  /**
   * Releases a connection and inform the connected node's listener by calling
   * notifyConnection(false). Closes any opened VerboseWriter file stream.
   */
  public void releaseConnection()
  {
    if (bc != null)
      bc.disconnect();
    if (bs != null)
      bs.close();
    VerboseWriter.close();
  }

  /**
   * Returns the Bluetooth friendly name of the local device.
   */
  public String getLocalName()
  {
    String localName = "";
    try
    {
      localName = LocalDevice.getLocalDevice().getFriendlyName();
    }
    catch (BluetoothStateException ex)
    {
    }
    return localName;
  }
}
