// AudioPlayer.java

package ch.aplu.gidlet;

import javax.microedition.media.*;
import java.io.*;

/**
 * Class that wraps the J2ME MMAPI class javax.microedition.media.Player in order
 * to start/resume, pause and stop playing audio files in a separate thread.
 */
public class AudioPlayer implements PlayerListener, Runnable
{
  interface State
  {
    int idle = 0;
    int start = 1;
    int pause = 2;
    int stop = 3;
    int close = 4;
  }

  private String mimeType;
  private InputStream is;
  private String uri = null;
  private volatile boolean isRunning = true;
  private Player player;
  private Thread t;
  private int state = State.idle;

  /**
   * Creates a Player in a separate thread from given resource file and mime type.
   * Blocks until Player's realize() and prefetch() are done.
   * Media time is zero.
   */
  public AudioPlayer(String audioFile, String mimeType)
  {
    is = getClass().getResourceAsStream(audioFile);
    this.mimeType = mimeType;
    init();
  }

  /**
   * Creates a Player in a separate thread from given input stream and mime type.
   * Blocks until Player's realize() and prefetch() are done.
   * Media time is zero.
   */
  public AudioPlayer(InputStream is, String mimeType)
  {
    this.is = is;
    this.mimeType = mimeType;
    init();
  }

  /**
   * Creates a Player in a separate thread from given HTTP uri.
   * Blocks until Player's realize() and prefetch() are done.
   * Media time is zero.
   */
  public AudioPlayer(String uri)
  {
    this.uri = uri;
    init();
  }

  private void init()
  {
    t = new Thread(this);
    t.start();
    // Wait until player's setup is done, otherwise player is
    // exposed by getPlayer() to early
     synchronized(this)
    {
      try
      {
        wait();
      }
      catch (InterruptedException ex) {}
    }
  }

  /**
   * Gets a reference to the Player instance.
   */
  public Player getPlayer()
  {
    return player;
  }

  /**
   * Starts or resumes playing at current media time.
   */
  public void start()
  {
    if (!t.isAlive())
    {
      Gidlet.showMessage("Error in AudioPlayer.start(): already closed", false);
      return;
    }
    synchronized(this)
    {
      state = State.start;
      notify();
    }
  }

  /**
   * Pauses playing. start() will resume at current media time.
   */
  public void pause()
  {
    if (!t.isAlive())
    {
      Gidlet.showMessage("Error in AudioPlayer.pause(): already closed", false);
      return;
    }
    synchronized(this)
    {
      state = State.pause;
      notify();
    }
  }

  /**
   * Stops playing and set media time to zero.
   */
  public void stop()
  {
    if (!t.isAlive())
    {
      Gidlet.showMessage("Error in AudioPlayer.stop(): already closed", false);
      return;
    }
    synchronized(this)
    {
      state = State.stop;
      notify();
    }
  }

  /**
   * Releases all connections and resources. No methods of this instance
   * should be called any more. Block until thread is dead.
   */
  public void close()
  {
    if (!t.isAlive())
      return;

    synchronized(this)
    {
      state = State.close;
      notify();
    }
    while (t.isAlive()) {}  // Wait until thread is dead
  }

  /**
   * For internal use only.
   */
  public void run()
  {
    try
    {
      if (uri == null)
        player = Manager.createPlayer(is, mimeType);
      else
        player = Manager.createPlayer(uri);
      if (Gidlet.isExiting())
      {
        player.close();
        return;
      }
      player.realize();
      if (Gidlet.isExiting())
      {
        player.close();
        return;
      }
      player.prefetch();
      if (Gidlet.isExiting())
      {
        player.close();
        return;
      }
      player.addPlayerListener(this);
      if (Gidlet.isExiting())
      {
        player.close();
        return;
      }
    }
    catch (Exception ex)
    {
      if (player != null) // In order to let the user treat this error via getPlayer()
        Gidlet.showMessage("Error while initializing player", false);
      isRunning = false;
    }

    synchronized(this)
    {
      notify();
    }

    while (isRunning && !Gidlet.isExiting())
    {
      synchronized(this)
      {
        try
        {
          wait();
        }
        catch (InterruptedException ex) {}
      }
      try
      {
        switch(state)
        {
          case State.start:
            player.start();
            break;

          case State.pause:
            player.stop();
            break;

          case State.stop:
            player.stop();
            player.setMediaTime(0);
            break;

          case State.close:
            player.close();
            isRunning = false;
            break;
        }
      }
      catch (MediaException ex)
      {
        Gidlet.showMessage("Error while sending commands to Player", false);
        isRunning = false;
      }
    }
  }


  /**
   * For internal use only.
   */
  public void playerUpdate(Player player, String event, Object eventData)
  {
    Gidlet.myGidlet.notifyPlayerUpdate(event, eventData);
  }
}
