// Gidlet.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
 */

package ch.aplu.gidlet;

import java.util.*;
import java.io.*;
import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;
import javax.wireless.messaging.*;
import javax.microedition.io.*;
import javax.microedition.media.*;


/**
 * Abstract helper class derived from MIDlet to provide functionalities
 * for starting and stopping the application. Typically the application class
 * extends this class. To simplify the application programming, some useful
 * methods and static variables are declared, so that no explicit class or instance
 * identifier is needed.<br><br>
 *
 * The platform's call of startApp() will create a separate thread (called "main-thread")
 * that calls main(). The application class should override main() and perform
 * all operations in the main() method.<br><br>
 * main() must terminate so that the main-thread stops running.<br><br>
 * If an endless loop in main() is needed, test with isExiting()
 * whether the Exit soft button has been hit. If the program blocks in the loop because
 * of waitOk(), readXXX(), showMessage(), main() will terminate immediately when
 * the Exit soft button is hit.
 * In this case an endless loop with while (true) is permitted.<br><br>
 *
 * Keep in mind that before main() runs, the environment is not fully initialized.
 * Therefore do not create visible GUI elements as instance variables. If needed
 * you may declare them as instance variable and create them in main(). The standard
 * Gidlet GUI elements starting with the letter 'M' (MForm, MConsole, etc.) may be
 * created as instance variable because there is an internal mechanism that defers
 * their initilization until main() runs.<br><br>
 *
 * When the platform requests the Gidlet to terminate by calling destroyApp(),
 * the Gidlet will terminate by calling doExit() that calls notifyDestroyed(), if
 * neither destroyApp() nor doExit() is overridden.<br><br>
 *
 * Before calling main() a special thread will be started in order disable the
 * the screen saver from shading the screen. If the screen saver should remain active,
 * declare a constructor and call enableScreenSaver(true).
 *
 */
public abstract class Gidlet extends MIDlet implements KeyListener, Runnable
{

 // --------------------- Inner class ReadThread ------------------
private class ReadThread extends Thread
{
  private StringBuffer sb;
  private InputStream is;

  private ReadThread(InputStream is, StringBuffer sb)
  {
    this.is = is;
    this.sb = sb;
  }

  public void run()
  {
    try
    {
      int ch;
      while ((ch = is.read()) != -1)
        sb.append((char)ch);

      is.close();
      isError = false;
    }
    catch (IOException ex)
    {
      try
      {
        is.close();
      }
      catch (Exception e) {}
      isError = true;
    }
  }
}


// --------------------- Inner class SMSReader ------------------
/**
 * Class that creates a SMS server thread based on the Wireless Messaging API (WMA).
 */
public class SMSReader extends Thread implements MessageListener
{
  private MessageConnection mc = null;
  private volatile boolean isRunning = true;
  private volatile int nbWaitingMsg = 0;

  /**
   * Creates a server thread based on the Wireless Messaging API (WMA)
   * that looks for incoming SMS.
   * The callback method notifyIncomingSMS() is triggered
   * at every received SMS.
   * You must specify a port number 0..65565 that corresponds to the
   * sender port number (SMS without a port number are captured by the
   * default SMS application and cannot be intercepted by a MIDlet).
   * To stop the thread, call stop().
   * @see #notifyIncomingSMS
   */
  public SMSReader(int portnumber)
  {
    if (portnumber < 0 || portnumber > 65535)
    {
      showMessage("Error in SMSReader: Illegal port number", false);
      return;
    }
    try
    {
      mc = (MessageConnection)Connector.open("sms://:" + portnumber);
      if (mc != null)
        mc.setMessageListener(this);
    }
    catch (IOException ex)
    {
      showMessage("Error in SMSReader: Can't open connection", false);
    }
    if (mc != null)
    {
      start();
    }
  }

  /**
   * For internal use only.
   */
  public void notifyIncomingMessage(MessageConnection mc)
  {
    if (mc == this.mc)
      takeMsg();
  }

  private synchronized void takeMsg()
  {
    nbWaitingMsg++;
    notify();
  }

  /**
   * Stops the SMSReader thread.
   */
  public void stop()
  {
    isRunning = false;
    if (mc != null)
    {
      try
      {
        mc.close();
        mc.setMessageListener(null);
      }
      catch (IOException ex) {}
    }
    interrupt();  // Stop wait()
  }

  /**
   * For internal use only.
   */
  public void run()
  {
    while (isRunning)
    {
      Message msg = null;
      synchronized(this)
      {
        if (nbWaitingMsg == 0)
        {
          try
          {
            wait();
          }
          catch (InterruptedException ex)
          {
            return;
          }
        }
        nbWaitingMsg--;
      }
      try
      {
         msg = mc.receive();
      }
      catch (IOException ex)
      {
        showMessage("Error while receiving SMS message", false);
        isRunning = false;
      }

      if (isRunning && msg instanceof TextMessage)
        notifyIncomingSMS((TextMessage)msg);
    }
  }
}


// --------------------- Inner class MMSReader ------------------
/**
 * Class that creates a MMS server thread based on the Wireless Messaging API (WMA).
 */
public class MMSReader extends Thread implements MessageListener
{
  private MessageConnection mc = null;
  private volatile boolean isRunning = true;
  private volatile int nbWaitingMsg = 0;

  /**
   * Creates a server thread based on the Wireless Messaging API (WMA)
   * that looks for incoming MMS.
   * The callback method notifyIncomingMMS() is triggered
   * at every received MMS.
   * You must specify an application identifier (appId) that corresponds to appId
   * of the sending MIDlet (MMS without an appId are captured by the
   * default MMS application and cannot be intercepted by a MIDlet).
   * To stop the thread, call stop().
   * @see #notifyIncomingSMS
   */
  public MMSReader(String appId)
  {
    try
    {
      mc = (MessageConnection)Connector.open("mms://:" + appId);
      if (mc != null)
        mc.setMessageListener(this);
    }
    catch (IOException ex)
    {
      showMessage("Error in MMSReader: Can't open connection", false);
    }
    if (mc != null)
    {
      start();
    }
  }

  /**
   * For internal use only.
   */
  public void notifyIncomingMessage(MessageConnection mc)
  {
    if (mc == this.mc)
      takeMsg();
  }

  private synchronized void takeMsg()
  {
    nbWaitingMsg++;
    notify();
  }

  /**
   * Stops the MMSReader thread.
   */
  public void stop()
  {
    isRunning = false;
    if (mc != null)
    {
      try
      {
        mc.close();
        mc.setMessageListener(null);
      }
      catch (IOException ex) {}
    }
    interrupt();  // Stop wait()
  }

  /**
   * For internal use only.
   */
  public void run()
  {
    while (isRunning)
    {
      Message msg = null;
      MultipartMessage mpm = null;
      Image receivedImage = null;
      synchronized(this)
      {
        if (nbWaitingMsg == 0)
        {
          try
          {
            wait();
          }
          catch (InterruptedException ex)
          {
            return;
          }
        }
        nbWaitingMsg--;
      }
      try
      {
         msg = mc.receive();
         if (msg instanceof MultipartMessage)
           mpm = (MultipartMessage)msg;
      }
      catch (IOException ex)
      {
        showMessage("Error while receiving SMS message", false);
        isRunning = false;
      }

      if (isRunning)
        notifyIncomingMMS(mpm);
    }
  }
}

// --------------------- Inner class HttpGetter ------------------
/**
 * Class that creates a thread to transfer data using the HTTP GET command.
 */
public class HttpGetter extends Thread
{
  private volatile boolean isRunning = true;

  private HttpConnection hc = null;
  private InputStream is = null;
  private int httpCode = 0;
  private String uri;
  private boolean isInternal;
  private ByteArrayOutputStream os = new ByteArrayOutputStream();

  /**
   * Creates a thread that transfers data from the given uri using
   * the HTTP GET command.
   * The callback method notifyHttpResponse() is triggered
   * when the transfer is completed. The thread is then terminated
   * and cannot be used any more.
   * @see #notifyHttpResponse
   */
  public HttpGetter(String uri)
  {
    isInternal = false;
    this.uri = uri;
    start();
  }

  // Used for internal notify only
  protected HttpGetter(String uri, boolean internalNotify)
  {
    this.isInternal = internalNotify;
    this.uri = uri;
    start();
  }

  /**
   * Stops the HttpGetter thread while transfering data.
   */
  public void stop()
  {
    isRunning = false;
  }

  /**
   * For internal use only.
   */
  public void run()
  {
    try
    {
      hc = (HttpConnection)Connector.open(uri);
    }
    catch (Exception ex)
    {
      if (isInternal)
       internalNotifyHttpResponse(null, -1);
      else
      {
        notifyHttpResponse(ex.getMessage().getBytes(), -1);
        notifyHttpResponse(ex.getMessage(), -1);
      }
      return;
    }

    try
    {
      httpCode = hc.getResponseCode();
      is = hc.openInputStream();
      byte[] data = new byte[4096];
      int nbBytes;
      while (isRunning && (nbBytes = is.read(data)) > 0)
      {
        os.write(data, 0, nbBytes);
      }
    }
    catch (Exception ex)
    {
      if (isInternal)
       internalNotifyHttpResponse(null, -1);
      else
      {
        notifyHttpResponse(ex.getMessage().getBytes(), httpCode);
        notifyHttpResponse(ex.getMessage(), httpCode);
      }
      return;
    }
    finally
    {
      try {is.close();} catch (Exception ex) {}  // catch also is = null
      try {hc.close();} catch (Exception ex) {}
    }
    byte[] data = os.toByteArray();
    if (isInternal)
      internalNotifyHttpResponse(data, httpCode);
    else
    {
      notifyHttpResponse(data, httpCode);
      notifyHttpResponse(new String(data), httpCode);
    }

    try {os.close();} catch (Exception ex) {}
  }
}



// --------------------- End of inner classes -------------------

  /**
   * List of all displays created before main() is called.
   */
  protected Vector displays = new Vector();

  /**
   * List of all turtles created before main() is called.
   */
  protected Vector turtles = new Vector();

  private Image transferedImage;
  private Displayable saveDisplay;
  private static Thread gidletThread;
  private  boolean isStarted = false;
  private  static boolean isExiting;
  private static Timer timer = new Timer();
  private static boolean isSleeping;
  private static long startTime;
  private static Object monitor = new Object();
  private static Thread t = null;
  private volatile boolean isError;
  private static boolean debug = false;

  /**
   * The current MIDlet's display.
   */
  public static Display display;

  /**
   * A reference to the current gidlet.
   */
  public static Gidlet myGidlet;


  /**
   * The color black.
   */
  public static final int BLACK = 0x000000;

  /**
   * The color blue.
   */
  public static final int BLUE = 0x0000FF;

  /**
   * The color cyan.
   */
  public static final int CYAN = 0x00FFFF;

  /**
   * The color darkgray.
   */
  public static final int DARKGRAY = 0x404040;

  /**
   * The color gray.
   */
  public static final int GRAY = 0x808080;

  /**
   * The color green.
   */
  public static final int GREEN = 0x00FF00;

  /**
   * The color lightgray.
   */
  public static final int LIGHTGRAY = 0xC0C0C0;

  /**
   * The color magenta.
   */
  public static final int MAGENTA = 0xFF00FF;

  /**
   * The color orange.
   */
  public static final int ORANGE = 0xFFC800;

  /**
   * The color pink.
   */
  public static final int PINK = 0xFFAFAF;

  /**
   * The color red.
   */
  public static final int RED = 0xFF0000;

  /**
   * The color white.
   */
  public static final int WHITE = 0xFFFFFF;

  /**
   * The color yellow.
   */
  public static final int YELLOW = 0xFFFF00;

  /**
   * Validation type "No validation"
   */
  public static final int NONE_VALIDATE = 0;

  /**
   * Validation type "Garantee to be an integer"
   */
  public static final int INT_VALIDATE = 1;

  /**
   * Validation type "Garantee to be a long"
   */
  public static final int LONG_VALIDATE = 2;

  /**
   * Validation type "Garantee to be a double"
   */
  public static final int DOUBLE_VALIDATE = 3;

  /**
   * Same as GameCanvas.FIRE_PRESSED.
   */
  public static final int  FIRE_PRESSED = GameCanvas.FIRE_PRESSED;

  /**
   * Same as GameCanvas.GAME_A_PRESSED.
   */
  public static final int  GAME_A_PRESSED = GameCanvas.GAME_A_PRESSED;

  /**
   * Same as GameCanvas.GAME_B_PRESSED.
   */
  public static final int  GAME_B_PRESSED = GameCanvas.GAME_B_PRESSED;

  /**
   * Same as GameCanvas.GAME_C_PRESSED.
   */
  public static final int  GAME_C_PRESSED = GameCanvas.GAME_C_PRESSED;

  /**
   * Same as GameCanvas.GAME_D_PRESSED.
   */
  public static final int  GAME_D_PRESSED = GameCanvas.GAME_D_PRESSED;

  /**
   * Same as GameCanvas.LEFT_PRESSED.
   */
  public static final int  LEFT_PRESSED = GameCanvas.LEFT_PRESSED;

  /**
   * Same as GameCanvas.RIGHT_PRESSED.
   */
  public static final int  RIGHT_PRESSED = GameCanvas.RIGHT_PRESSED;

  /**
   * Same as GameCanvas.UP_PRESSED.
   */
  public static final int  UP_PRESSED = GameCanvas.UP_PRESSED;

  /**
   * Same as GameCanvas.DOWN_PRESSED.
   */
  public static final int DOWN_PRESSED = GameCanvas.DOWN_PRESSED;

  /**
   * Constructs a Gidlet instance and perform a few initializations.
   */
  public Gidlet()
  {
    myGidlet = this;  // Not yet fully constructed, don't use it before startApp()
  }

  /**
   * Implementation of MIDlet.startApp(). If called the first time, create a
   * new thread and run main(). If called after a pause, the display stored
   * at this is moment is reshown.
   */
  protected void startApp()
  {
    if (isStarted)
    {
      display.setCurrent(saveDisplay);
    }
    else
    {
      isExiting = false;
      isSleeping = false;
      isStarted = true;
      display = Display.getDisplay(this);
      gidletThread = new Thread(this, "Gidlet-Thread");
      gidletThread.start();
    }
  }

  /**
   * Returns the main thread.
   */
  public static Thread getThread()
  {
    return gidletThread;
  }

  /**
   * Abstract declaration, must be overridden by derived class.
   */
  abstract public void main();

  /**
   * Called when the platform requests to pause the program.
   * Will call doPause(), which is empty but may be
   * overriden by a user's callback.
   */
  protected void pauseApp()
  {
    saveDisplay = display.getCurrent();
    doPause();
  }

  /**
   * Called when the platform requests to terminate the program.
   * Will call doExit(), which calls notifyDestroyed(), if not
   * overriden by a user's callback.
   */
  protected void destroyApp(boolean b)
  {
    doExit();
  }

  /**
   * Empty implementation of KeyListener.keyPressed().
   */
  public void keyPressed(int keyCode) {}

  /**
   * Empty implementation of KeyListener.keyReleased().
   */
  public void keyReleased(int keyCode) {}

  /**
   * Empty implementation of KeyListener.keyRepeated().
   */
  public void keyRepeated(int keyCode) {}

  /**
   * Gidlet's run-method implementation. Do not override.
   */
  public void run()
  {
    for (int i = 0; i < displays.size(); i++)
    {
      Object obj = displays.elementAt(i);
      if (obj instanceof MPanel)
        ((MPanel)obj).init();
      if (obj instanceof MConsole)
        ((MConsole)obj).init();
      if (obj instanceof Playground)
        ((Playground)obj).init();
    }

    for (int i = 0; i < turtles.size(); i++)
    {
      Object obj = turtles.elementAt(i);
      ((Turtle)obj).init();
    }

    // RuntimeException thrown when exiting, must terminate thread
    try
    {
      main();
    }
    catch (RuntimeException ex) {}
    if (debug)
      showDebug("Debug:::main thread terminated");
  }

  /**
   * Terminates the application by calling wakeUp() and notifyDestroyed(). Called when the Exit
   * soft button is hit. Also called by destroyApp(), when the system requests to terminate.
   * wakeUp() is called to unblock the main-thread if it is in a waiting state.<br>
   * Override to get the callback and run your cleanup code. Your callback method
   * is then responsible to unblock a waiting main-thread and call notifyDestroyed().<br><br>
   * If the main-thread is blocked in waitOk(), you may call super.doExit()
   * to terminate the main-thread.
   */
  protected void doExit()
  {
    if (debug)
      showDebug("Debug:::doExit() entered");
    isExiting = true;
    wakeUp();
    notifyDestroyed();
  }

  /**
   * Empty. Override to get a callback from the Midlet's pause event.
   */
  protected void doPause() {}

  /**
   * Empty. Override to get a callback, when the OK soft button is hit.
   * Call the display's addOkButton() to show the button.
   * As usual, keep in mind that the callback method must return quickly,
   * particularly do not create other displays within this callback. Because
   * showMessage() and the readXXX() methods create a Form instance, do not invoke them
   * in within doOk().
   * If several displays are used, the given Displayable can be used, to distinguish
   * between the different OK labels. <br>
   * (Remind that Displayable is a superclass of MConsole, MForm and MPanel.)
   */


   void doOk(Displayable d)  {}

  /**
   * Empty. Override to get a callback, when the OK soft button is hit.
   * Call the display's addOkButton() to show the button.
   * As usual, keep in mind that the callback method must return quickly,
   * particularly do not create other displays within this callback. Because
   * showMessage() and the readXXX() methods create a Form instance, do not invoke
   * them within doOk().<br>
   */
  protected void doOk() {}


  /**
   * Waits until a OK soft button is hit.
   * If addOkButton = true and the current display is one of MForm,
   * MTextBox, MPanel, MList, MConsole, Playground,
   * the OK soft button is automatically shown and removed
   * after the OK button is hit.<br>
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main() unless you override doExit(). In this case you should call super.doExit()
   * to terminate the application.
   */
  protected static void waitOk(boolean addOkButton)
  {
    if (isExiting)
      return;

    MForm f = null;
    MConsole c = null;
    MList l = null;
    MPanel p = null;
    MTextBox t = null;
    Playground pg =  null;
    Displayable d = null;

    if (addOkButton)
    {
      d = display.getCurrent();

      if (d instanceof MForm)
      {
        f = (MForm)d;
        f.addOkButton();
      }
      if (d instanceof MConsole)
      {
        c = (MConsole)d;
        c.addOkButton();
      }
      if (d instanceof MList)
      {
        l = (MList)d;
        l.addOkButton();
      }
      if (d instanceof MPanel)
      {
        p = (MPanel)d;
        p.addOkButton();
      }
      if (d instanceof MTextBox)
      {
        t = (MTextBox)d;
        t.addOkButton();
      }
      if (d instanceof Playground)
      {
        pg = (Playground)d;
        pg.addOkButton();
      }
    }

    isSleeping = true;
    putSleep();
    // Can only throw exception in same thread
    if (isExiting && Thread.currentThread() == gidletThread)
      throw new RuntimeException();  // Terminate main()

    if (addOkButton)
    {
      if (f != null)
        f.removeOkButton();
      if (c != null)
        c.removeOkButton();
      if (l != null)
        l.removeOkButton();
      if (p != null)
        p.removeOkButton();
      if (t != null)
        t.removeOkButton();
      if (pg != null)
        pg.removeOkButton();
    }
  }

  /**
   * Waits until the OK soft button is hit.
   * If the current display is one of MForm, MTextBox, MPanel,
   * MList, MConsole, Playground, the OK soft button is
   * automatically shown and removed after the OK button is hit.<br>
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main() unless you override doExit(). In this case you should call super.doExit()
   * to terminate the application.
   */
  protected static void waitOk()
  {
    waitOk(true);
  }

  /**
   * Shows a simple form with an input field with the given prompt message
   * and given initial value. Only non empty integer values are accepted.
   * Block until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  protected static int readInt(String prompt, int init)
  {
    return internalReadInt(prompt, new Integer(init).toString());
  }

  /**
   * Shows a simple form with an input field with the given prompt message
   * and no initial value. Only non empty integer values are accepted.
   * Block until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  protected static int readInt(String prompt)
  {
    return internalReadInt(prompt, "");
  }

  private static int internalReadInt(String prompt, String init)
  {
    Displayable currentDisplay = display.getCurrent();
    MForm form = new MForm("OK to continue.");
    Item it = form.addInputField(prompt, init, INT_VALIDATE);
    form.addOkButton();
    form.show();
    String result;
    do
    {
      waitOk(false);
      result = form.getText(it);
    } while (result.equals(""));
    display.setCurrent(currentDisplay);
    return Integer.parseInt(result);
  }

  /**
   * Shows a simple form with an input field with the given prompt message
   * and given initial value. Only non empty integer values are accepted.
   * Blocks until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  protected static long readLong(String prompt, long init)
  {
    return internalReadLong(prompt, new Long(init).toString());
  }

  /**
   * Shows a simple form with an input field with the given prompt message
   * and no initial value. Only non empty integer values are accepted.
   * Blocks until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  protected static long readLong(String prompt)
  {
    return internalReadLong(prompt, "");
  }

  private static long internalReadLong(String prompt, String init)
  {
    Displayable currentDisplay = display.getCurrent();
    MForm form = new MForm("OK to continue.");
    Item it = form.addInputField(prompt, init, LONG_VALIDATE);
    form.addOkButton();
    form.show();
    String result;
    do
    {
      waitOk(false);
      result = form.getText(it);
    } while (result.equals(""));
    display.setCurrent(currentDisplay);
    return Long.parseLong(result);
  }

  /**
   * Shows a simple form with an input field with the given prompt message,
   * and given initial value. Only non empty double values are accepted.
   * Blocks until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
      */
  protected static double readDouble(String prompt, double init)
  {
    return internalReadDouble(prompt, new Double(init).toString());
  }

  /**
   * Shows a simple form with an input field with the given prompt message
   * and no initial value. Only non empty double values are accepted.
   * Blocks until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  protected static double readDouble(String prompt)
  {
    return internalReadDouble(prompt, "");
  }

  private static double internalReadDouble(String prompt, String init)
  {
    Displayable currentDisplay = display.getCurrent();
    MForm form = new MForm("OK to continue.");
    Item it = form.addInputField(prompt, init, DOUBLE_VALIDATE);
    form.addOkButton();
    form.show();
    String result;
    do
    {
      waitOk(false);
      result = form.getText(it);
    } while (result.equals(""));
    display.setCurrent(currentDisplay);
    return Double.parseDouble(result);
  }

  /**
   * Shows a simple form with an input field with the given prompt message,
   * and given initial value. Any character sequence is accepted.
   * Blocks until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  protected static String readString(String prompt, String init)
  {
    return internalReadString(prompt, init);
  }

  /**
   * Shows a simple form with an input field with the given prompt message
   * and no initial value. Any character sequences is accepted.
   * Blocks until the OK button is pressed and returns to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  protected static String readString(String prompt)
  {
    return internalReadString(prompt, "");
  }

  private static String internalReadString(String prompt, String init)
  {
    Displayable currentDisplay = display.getCurrent();
    MForm form = new MForm("OK to continue.");
    Item it = form.addInputField(prompt, init, NONE_VALIDATE);
    form.addOkButton();
    form.show();
    waitOk(false);
    display.setCurrent(currentDisplay);
    return form.getText(it);
  }

  /**
   * Shows a simple form with an output field that contains
   * the given message.
   * Blocks until the OK button is pressed and return to the previous display,
   * if restore = true.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  public static void showMessage(String msg, boolean restore)
  {
    Displayable currentDisplay = display.getCurrent();
    // May be null, if no current display is yet selected

    String title;
    if (restore)
      title = "OK to continue";
    else
      title = "Exit to terminate";

    MForm form = new MForm(title);
    if (restore)
      form.addOkButton();
    StringItem si = form.addOutputField();
    form.setText(si, msg);
    form.show(true);
    waitOk(false);
    if (restore)
    {
      form.setText(si, ""); // Clear all, because we never use it again.
                            // It should not be redisplayed if a
                            // long calculation follows
      if (currentDisplay != null)
      {
        display.setCurrent(currentDisplay);
        while (!currentDisplay.isShown())  // Wait until really shown
          Thread.yield();
      }
    }
  }

  /**
   * Shows a simple form with an output field that contains
   * the given message.
   * Blocks until the OK button is pressed and return to the previous display.
   * If the Exit soft button is hit, a RuntimeException is thrown that terminates
   * main().
   */
  public static void showMessage(String msg)
  {
    showMessage(msg, true);
  }

  /**
   * Shows a simple form during the given duration (in ms) with an output field that contains
   * the given message.
   * Returns immediately, but callback-method messageTimeout() will be triggerd
   * when duration elapses. By overriding messageTimeout() you get your own notification.
   * The callback is canceled by calling closeTimedMessage().<br>
   * If duration <= 0, no callback is triggered.
   */
  protected void showTimedMessage(String msg, int duration)
  {
    MForm form = new MForm();
    StringItem si = form.addOutputField();
    form.setText(si, msg);
    form.show(true);
    if (duration > 0)
    {
      timer.schedule(new TimerTask()
      {
        public void run()
        {
          messageTimeout();
        }
      }, duration);
    }
  }

  /**
   * Will be called when the timeout of the timed message expires.
   * Empty, should be overridden to get the notification.
   */
  protected void messageTimeout()
  {}

  /**
   * Closes a timed message box before timeout expires.
   * messageTimeout() will not be called.
   */
  protected void closeTimedMessage()
  {
    timer.cancel();
  }

  /**
   * Reads and returns text from a textfile with given URL.
   * Forward slashes must be used as directory separators.
   * Return null, if textfile isn't found.
   */
   public static String readTextResource(String fileUrl)
   {
     String object = new String();  // Trick to be able to use getClass in a static context
     InputStream is = object.getClass().getResourceAsStream(fileUrl);
     String text = null;
     if (is == null)
       return null;

     try
     {
       StringBuffer sb = new StringBuffer();
       InputStreamReader isr = new InputStreamReader(is);
       int i;
       while ((i = isr.read()) != -1)
         sb.append((char)i);
       isr.close();
       text = sb.toString();
     }
     catch (IOException ex)
     {
       return null;
     }
     return text;
   }

  /**
   * Loads image from given resource and returns it's Image reference.
   * Forward slashes must be used as directory separators.
   * Returns null, if loading fails.
   */
  public static Image getImage(String imageUrl)
  {
    String object = new String(); // Trick to be able to use getClass in a static context
    InputStream is = object.getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);  // Blocking until read and decoded
      if (is != null)
        is.close();
    }
    catch (IOException ex)
    {
      return null;
    }
    return img;
  }

  /**
   * Loads resource into byte array.
   * Forward slashes must be used as directory separators.
   * Returns null, if loading fails.
   */
  public static byte[] getResource(String resourceUrl)
  {
    InputStream is = null;
    byte[] data = null;
    try
    {
      String object = new String(); // Trick to be able to use getClass in a static context
      is = object.getClass().getResourceAsStream(resourceUrl);
      data = new byte[is.available()];
      is.read(data);
    }
    catch (Exception ex)
    {
      return null;
    }
    finally
    {
      try {is.close();} catch (Exception ex) {}
    }
    return data;
  }

  /**
   * Loads image from given HTTP uri and returns it's Image reference. Blocks
   * until image is loaded or time expires the given timeout (in ms).
   * Returns the loaded image or null, if loading fails.
   */
  public Image getImageHttp(String imageUri, int timeout)
  {
    HttpGetter hg = new HttpGetter(imageUri, true);  // Wait for internalNotify
    boolean rc = putSleep(timeout);
    if (!rc) // timeout
    {
      hg.stop();
      return null;
    }
    // Got notification
    return transferedImage; // May still be null
  }

  protected void internalNotifyHttpResponse(byte[] data, int code)
  {
    int okStatusCode = 200;
    if (code != okStatusCode)
      transferedImage = null;
    else
      transferedImage = Image.createImage(data, 0, data.length);
    wakeUp();
  }

  /**
   * Loads image from given resource and returns it's Sprite reference.
   * Returns null, if loading fails.<br>
   */
  public static Sprite getSprite(String imageUrl)
  {
    Image spriteImg = getImage(imageUrl);
    if (spriteImg == null)
      return null;
    else
      return new Sprite(spriteImg);
  }

  /**
   * Loads image from given resource with given frameWidth, frameHeight
   * and returns it's Sprite reference.
   * Returns null, if loading fails.<br>
   */
  public static Sprite getSprite(String imageUrl, int frameWidth, int frameHeight)
  {
    Image spriteImg = getImage(imageUrl);
    if (spriteImg == null)
      return null;
    else
      return new Sprite(spriteImg, frameWidth, frameHeight);
  }

  /**
   * Sends a SMS with given text to given phone and SMS port number.
   * Blocks until the message is sent.
   * If portnumber <= 0, no port number is used.
   * Return values:<br>
   * 0: successful <br>
   * 1: text too long
   * 2: unable to send to the given phone number<br>
   */
  protected static int sendSMS(String phonenumber, int portnumber, String text)
  {
    MessageConnection mc = null;;
    try
    {
      String smsUrl = "sms://" + phonenumber;
      if (portnumber > 0)
         smsUrl += ":" + portnumber;
      mc = (MessageConnection)Connector.open(smsUrl);
      TextMessage tm =
        (TextMessage)mc.newMessage(MessageConnection.TEXT_MESSAGE);
      tm.setPayloadText(text);
      if (mc.numberOfSegments(tm) == 0)
        return 1;
      mc.send(tm);
    }
    catch (IOException ex)
    {
      return 2;
    }
    finally
    {
      try {mc.close();} catch (Exception ex) {};
    }
    return 0;
  }

  /**
   * Sends a SMS with given text to given phone number (and no SMS port number).
   * Blocks until the message is sent.
   * Return values:<br>
   * 0: successful <br>
   * 1: text too long
   * 2: unable to send to the given phone number<br>
   */
  protected static int sendSMS(String phonenumber, String text)
  {
    return sendSMS(phonenumber, 0, text);
  }

  /**
   * Empty callback method triggered by the SMSReader thread when a SMS is received.
   * Override and use
   *   msg.getAddress():     String that contains the sender's identification
   *   msg.getTimestamp():   java.util.Date that contains the time of arrival
   *   msg.getPayloadText(): String that contains the message
   * Keep in mind that callback method must return quickly. Avoid calls to
   * blocking method.
   * @see SMSReader
   */
  public void notifyIncomingSMS(TextMessage msg)
  {}

  /**
   * Sends a MMS to given phone number with given application identifier, subject,
   * message parts and startContentId.
   * If appId = null, no application identifier is used and the MMS is sent
   * to the default MMS application (and not to a MIDlet).
   * The startContendId must be one of the ids of the MessageParts.
   * It helps the receiver to select the part the is
   * responsible for presentation. For more information consult the WMA documentation.
   * Spawns a thread for doing the job and returns immediately. When the job
   * is done, notifyOutgoingMMS() is triggered.
   * @see #notifyOutgoingMMS
   */
  protected static void sendMMS(final String phonenumber, final String appId,
                            final String subject, final MessagePart[] parts,
                            final String startContentId)
  {
      new Thread()
        {
          public void run()
          {
            int error = -1;
            MessageConnection mc =  null;
            String mmsUrl;
            try
            {
              if (appId == null)
                mmsUrl = "mms://" + phonenumber;
              else
                mmsUrl = "mms://" + phonenumber + ":" + appId;
              mc = (MessageConnection)Connector.open(mmsUrl);
              MultipartMessage mpm =
                (MultipartMessage)mc.newMessage(MessageConnection.MULTIPART_MESSAGE);
              mpm.setSubject(subject);
              for (int i = 0; i < parts.length; i++)
                mpm.addMessagePart(parts[i]);
              if (startContentId != null)
                mpm.setStartContentId(startContentId);
              mc.send(mpm);
            }
            catch (IOException ex)
            {
              error = 1;
            }
            finally
            {
              try {mc.close();} catch (Exception ex){};
            }
            error = 0;
            myGidlet.notifyOutgoingMMS(error);
          }
        }.start();
  }

  /**
   * Sends a MMS to given phone number with given application identifier, subject,
   * and message parts.
   * If appId = null, no application identifier is used and the MMS is sent
   * to the default MMS application (and not to a MIDlet).
   * No startContentId is used.
   * For more information consult the WMA documentation.
   * Spawns a thread for doing the job and returns immediately. When the job
   * is done, notifyOutgoingMMS() is triggered.
   * @see #notifyOutgoingMMS
   */
  protected static void sendMMS(String phonenumber, String appId, String subject,
                            MessagePart[] parts)
  {
    sendMMS(phonenumber, appId, subject, parts, null);
  }

  /**
   * Sends an e-mail via MMS with given recipient's e-mail address, subject and message.
   * The MMS MessagePart's encoding scheme is "ISO-8859-1".
   * Returns immediately, but notifyOutgoingMMS() is triggered when job is done.
   * @see #notifyOutgoingMMS
   */
  protected static void sendMMSMail(String recipient, String subject, String message)
  {
    try
    {
      MessagePart textPart =
        new MessagePart(message.getBytes(), "text/plain", "id", "location", "ISO-8859-1");
      MessagePart[] parts = {textPart};
      sendMMS(recipient, null, subject, parts);
    }
    catch (SizeExceededException ex)
    {
      showMessage("Error in sendMMSMail. Message too big", false);
    }
  }

  /**
   * Empty callback method triggered by the sendMMS() and sendMMSMail() thread
   * when it terminates.
   * Keep in mind that callback method must return quickly. Avoid calls to
   * blocking method.<br>
   * status values:<br>
   * 0: success<br>
   * 1: error<br>
   * @see #sendMMS
   * @see #sendMMSMail
   */
  public void notifyOutgoingMMS(int status)
  {}

  /**
   * Empty callback method triggered by the MMSReader thread when a MMS is received.
   * Override and use  mpm.getMessagePart(String contentID).getContent()
   * to retrieve the data from message parts.
   * For more information consult the WMA documentation.
   * Keep in mind that callback method must return quickly. Avoid calls to
   * blocking method.
   * @see MMSReader
   */
  public void notifyIncomingMMS(MultipartMessage mpm)
  {}

  /**
   * Empty callback method triggered by HttpReader when the transfer due to the
   * HTTP GET command issued by the HttpGetter constructor terminates or
   * when an error occured. Mainly used for transfering binary files (e.g. gif-images).
   * Override and use
   *   response:  data returned from server in reponse to the HTTP GET command
   *   httpCode:  HTTP status returned from server or -1 if the request fails
   *      (in the latter case, response contains the error description)<br>
   * Response contains the binary data.
   * Conversion to a string may be performed with <code>new String(data)</code>
   * Keep in mind that callback method must return quickly. Avoid calls to
   * blocking methods.
   * @see HttpGetter
   */
  public void notifyHttpResponse(byte[] response, int httpCode)
  {}

  /**
   * Empty callback method triggered by HttpReader when the transfer due to the
   * HTTP GET command issued by the HttpGetter constructor terminates or
   * when an error occured.
   * Override and use
   *   response:  data returned from server in reponse to the HTTP GET command
   *   httpCode:  HTTP status returned from server or -1 if the request fails
   *      (in the latter case, response contains the error description)<br>
   * Response contains the text content of the Web page.
   * Keep in mind that callback method must return quickly. Avoid calls to
   * blocking methods.
   * @see HttpGetter
   */
  public void notifyHttpResponse(String response, int httpCode)
  {}

  /**
   * Plays a system sound of given type.<br>
   * 0: INFO
   * 1: CONFIRMATION
   * 2: WARNING
   * 3: ALARM
   * 4: ERROR
   * The sound is only emitted if a display (Form, MPanel, etc.) is shown.
   */
  public static void beep(int type)
  {
    switch (type)
    {
      default:
      case 0:
        AlertType.INFO.playSound(display);
        break;

      case 1:
        AlertType.CONFIRMATION.playSound(display);
        break;

      case 2:
        AlertType.WARNING.playSound(display);
        break;

      case 3:
        AlertType.ALARM.playSound(display);
        break;

      case 4:
        AlertType.ERROR.playSound(display);
        break;
    }
  }

  /**
   * Returns version information of Gidlet framework.
   */
  public static String getVersion()
  {
    return SharedConstants.VERSION;
  }

  /**
   * Returns a reference to the current gidlet instance.
   */
  public static Gidlet getGidlet()
  {
    return myGidlet;
  }

  /**
   * Starts a timer or restart it by setting its time to zero.
   */
  public static void startTimer()
  {
    startTime = System.currentTimeMillis();
  }

  /**
   * Gets timer's time (in ms).
   */
  public static long getTime()
  {
    if (startTime == 0)
      return 0L;
    else
      return System.currentTimeMillis() - startTime;
  }

  /**
   * Delays execution for the given amount of time (in ms).
   * If time is less or equal zero, return immediately.
   */
  public static void delay(int time)
  {
    if (time <= 0)
      return;
    try
    {
      Thread.sleep(time);
    }
    catch (Exception e) {}
  }

  /**
   * Puts the current thread in a wait state until wakeUp() is called
   * or timeout (in ms) expires. If timeout <= 0 the method blocks infinitely
   * until wakeUp() is called. Only one thread may be in the wait state.<br>
   * Return values:<br>
   * true:  wakeUp was called before timeout expired ("no timeout occured")<br>
   * false: timeout expired before wakeUp was called ("timeout occured")<br>
   */
  public static boolean putSleep(final int timeout)
  {
    final Thread currentThread = Thread.currentThread();
    if (timeout > 0)
    {
      t = new Thread()  // Timeout thread
      {
        public void run()
        {
          boolean rc = waitTimeout(timeout);
          if (rc)
            currentThread.interrupt();
        }
      };
      t.start();
    }

    isSleeping = true;
    synchronized(monitor)
    {
      try
      {
        monitor.wait();
      }
      catch (InterruptedException ex)
      {
        return false;
      }
    }
    return true;
  }

  /**
   * Same as putSleep(int timeout) with timeout = 0 (timeout disalbed).
   */
  public static boolean putSleep()
  {
    return putSleep(0);
  }

  /**
   * Wakes up the waiting thread.
   */
  public static void wakeUp()
  {
    if (t != null)
      t.interrupt();  // Stop delay(), this will stop the timeout thread
    synchronized(monitor)
    {
      monitor.notify();
    }
    isSleeping = false;
  }

  private static boolean waitTimeout(int time)
  {
    try
    {
      t.sleep(time);
    }
    catch (InterruptedException ex)
    {
      return false;
    }
    return true;
  }

  /**
   * Returns true, if putSleep() was called and still waiting for wakeUp() or
   * timeout.
   * @see #putSleep
   * @see #wakeUp
   */
  public static boolean isSleeping()
  {
    return isSleeping;
  }

  /**
   * Empty callback method triggered by the PlayerListener when
   * the state of the Player is updated. See MMAPI documentation
   * of class javax.microedition.media.PlayerListener for information
   * about event and eventData.
   * @see javax.microedition.media.PlayerListener
   */
  public void notifyPlayerUpdate(String event, Object eventData)
  {}

  /**
   * Returns true, if the application runs in the emulator, otherwise false.
   * This methods checks if the system property with key "microedition.platform"
   * ends with "wtk". This may fail on some platforms.
   */
  public static boolean isEmulator()
  {
    return System.getProperty ("microedition.platform").endsWith ("wtk");
  }

  /**
   * Returns true, if the Exit soft button is hit in a readXXX(),
   * showMessage() or showTimeMessage() method.
   */
  public static boolean isExiting()
  {
    return isExiting;
  }

  /**
   * Turns on/off debug information on System.out (default: off).<br>
   * The message is displayed on System.out, but may be redirected by
   * overriding showDebug().
   * @see #showDebug
   */
  public static void setDebug(boolean debug)
  {
    Gidlet.debug = debug;
  }

  /**
   * Returns true if debug information is turned on.
   */
  public static boolean getDebug()
  {
    return debug;
  }

  /**
   * Splits given string with given separator string into multiple strings.
   * Returns splitted string array (without separator).<br>
   * Array has length 0, if input string is null or empty.<br>
   * If the given string terminates with the separator,
   * the last array element is an empty string.
   */
  public static String[] split(String s, String separator)
  {
    if (s == null || s.equals(""))
      return new String[0];

    Vector nodes = new Vector();

    // Parse nodes into vector
    int index = s.indexOf(separator);
    while (index >= 0)
    {
      nodes.addElement(s.substring(0, index));
      s = s.substring(index + separator.length());
      index = s.indexOf(separator);
    }

    // Get the last node
    nodes.addElement(s);

    // Create splitted string array
    String[] result = new String[nodes.size()];
    if (nodes.size() > 0)
    {
      for (int i = 0; i < nodes.size(); i++)
        result[i] = (String)nodes.elementAt(i);
    }
    return result;
  }

  /**
   * Reads content of a text file from a given url.
   * Runs in a separate thread to improve performance,
   * but waits for the thread to terminate.
   * All resources are released when finishing.
   * @return content as string, empty if file has zero length, null if operation fails
   */
  public String readFile(String fileUrl)
  {
    InputStream is = getClass().getResourceAsStream(fileUrl);
    if (is == null)
      return null;

    StringBuffer sb = new StringBuffer();
    ReadThread rt = new ReadThread(is, sb);
    rt.start();
    try
    {
      rt.join();
    }
    catch (InterruptedException ex) {}
    if (isError)
      return null;
    return sb.toString();
  }

  /**
   * Displays given debug message on System.out, when debugging is enabled by
   * setDebug(true).<br>
   * May be overridden to redirect the message.
   * @see #setDebug
   */
  protected void showDebug(String msg)
  {
    System.out.println(msg);
  }
}

