// MConsole.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
*/

/* Fundamental ideas from Nokia's  - Wireless Software Solutions */

package ch.aplu.gidlet;

import java.util.*;
import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import ch.aplu.gidlet.*;

/**
 * Full screen text window containing lines of text, simulating Java's
 * System.out.println.<br><br>
 * Each line may have an individual color. No text cursor is shown.
 * If the number of displayed lines exceeds the height of the display, the
 * text may be vertically scrolled by using the up/down keys and
 * horizontally scrolled by using the left/right keys.<br><br>
 * Newline character \n may be inserted to start a new line,
 * other non-printing characters should not be used.<br><br>
 * More than one instance may be created. The constructor flag visible
 * may be used, to select if the display is initially shown or hidden.
 * A hidden display will become visible by calling show(). There is no other way to
 * make it hidden than displaying a different display.
 * print/println may be called while the display is hidden.<br><br>
 *
 * A soft button 'Exit' is shown. Hitting the corresponding key will
 * call the Gidlet's doExit(). If you do not need this button,
 * invoke removeExitButton().<br><br>
 * If you implement your own CommandListener and call setCommandListener(), the internal
 * CommandListener is disabled.<br><br>
 *
 * print() and println() are overloaded for all primitive data types.<br><br>
 *
 * main() serves as entry point for the user programm. It is called in the run-method
 * of a separate thread and should terminate to avoid non-terminated threads.<br><br>
 *
 * A global instance of MConsole may be created while declaring application instance variables.
 * To garantee that the initialization is executed in the Gidlet's thread,
 * these instances are queued and the intialisation takes place in the same order
 * they are created when main() is started.
 */
public class MConsole extends Canvas
{
// ------------------ Inner class MyCommandListener -------------
class MyCommandListener implements CommandListener
{
  public void commandAction(Command c, Displayable d)
  {
    if (c == exitCommand)
    {
      keyThread.interrupt();
      try
      {
        keyThread.join();  // Wait until dead
      }
      catch (InterruptedException ex) {}
      gidlet.doExit();
    }

    if (okCommand != null && c == okCommand)
    {
      gidlet.doOk();
      gidlet.doOk(d);
      if (gidlet.isSleeping())
        gidlet.wakeUp();
    }
  }
}

// ------------------ Inner class TimerTask ---------------------
class MyTimerTask extends TimerTask
{
  public void run()
  {
    waitVisible = false;
  }
}

// ------------------ Inner class KeyThread ---------------------
class KeyThread extends Thread
{
  public void run()
  {
    while (true)
    {
      synchronized(keyThread)
      {
        try
        {
          keyThread.wait();
        }
        catch (InterruptedException ex)
        {
          return;
        }
      }
      if (action == UP && startIndex > 0)
      {
         startIndex--;
         repaint();
      }

      if (action == DOWN && lines.size() > displayLines
                         && startIndex + displayLines < lines.size())
      {
        startIndex++;
        repaint();
      }

      if (action == LEFT && x < 0)
      {
        x += HSTEP;
        repaint();
      }

      if (action == RIGHT)
      {
        x -= HSTEP;
        repaint();
      }
    }
  }
}
// ------------------ End of inner classes ----------------------

  private Gidlet gidlet;
  private Thread keyThread;
  private Command exitCommand = new Command("Exit", Command.EXIT, 1);
  private Command okCommand = null;
  private static final int ANCHOR = Graphics.TOP | Graphics.LEFT;
  private static final int HSTEP  = 7;
  private static final int defaultTextColor = Gidlet.WHITE;
  private static final int defaultBgColor = Gidlet.BLACK;

  private Vector lines;
  private Vector colors;
  private Font font;
  private int displayLines;
  private int startIndex;
  private int nbLines;
  private int action;
  private int x;
  private int bgColor;
  private int fSize;
  private int textColor;
  private boolean keepLine = false;
  private boolean waitVisible;
  private int fontSize;
  private boolean visible;

  // ============== Constructors =========================================

  /**
   * Constructs a MConsole with given maximal number of lines in text buffer,
   * the background color, the default text color and
   * the size of the font (1: small, 2: medium, 3: large)<br>
   * If nbLines <= 0, the number of lines is unlimited.<br>
   * The parameter visible controls, if the display is initially shown.
   */
  public MConsole(int nbLines, int bgColor, int textColor, int fontSize, boolean visible)
  {
    super();
    this.gidlet = Gidlet.myGidlet;
    if (nbLines < 0)
      nbLines = 0;
    this.nbLines = nbLines;
    this.bgColor = bgColor;
    this.textColor = textColor;
    this.fontSize = fontSize;
    this.visible = visible;
    if (Thread.currentThread().getName().equals("Gidlet-Thread"))
      init();
    else // defer initialization
      (gidlet.displays).addElement(this);
    }

  /**
   * Constructs a MConsole with given maximal number of lines in text buffer,
   * the background color, the default text color and
   * the size of the font (1: small, 2: medium, 3: large)<br>
   * If nbLines <= 0, the number of lines is unlimited.<br>
   * The display is initially shown.
   */
  public MConsole(int nbLines, int bgColor, int textColor, int fontSize)
  {
    this(nbLines, bgColor, textColor, fontSize, true);
  }

  /**
   * Contructs a MConsole with unlimited number of lines, black background,
   * white default text color and large size of font.
   * The parameter visible controls, if the display is initially shown.
   */
  public MConsole(boolean visible)
  {
    this(0, defaultBgColor, defaultTextColor, 3, visible);
  }

  /**
   * Contructs a MConsole with unlimited number of lines, black background,
   * white default text color and large size of font.
   * The display is initially shown.
   */
  public MConsole()
  {
    this(true);
  }

  // ============== End of constructors ===================================

  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * Because it is not garanteed that the diplay is actually visible when
   * Display.setCurrent() returns, wait for a maximum of 5 s for the display
   * to get shown.
   *
   */
   public void show()
   {
     show(true);
   }

   /**
    * Shows the console display if it was previously hidden by another display
    * or not yet displayed.
    * If wait = true, block until the display becomes actually visible, up
    * to a maximum auf 5 seconds. This may be useful to observe the start of
    * a calculation.
    */
   public void show(boolean wait)
   {
     gidlet.display.setCurrent(this);
     if (wait)
     {
       waitVisible = true;
       Timer t = new Timer();
       t.schedule(new MyTimerTask(), 5000);
       while (waitVisible && !isShown())
         Gidlet.delay(100);
       t.cancel();
     }
   }


  /**
   * Changes default text color to given color. The change will only become
   * effective on the next line.
   */
   public void color(int textColor)
   {
     this.textColor = textColor;
   }

  /**
   * Appends given text to last line, if print() was previously called or
   * start a new line, if no print-method or println() was previously called.
   */
  public synchronized void println(String text)
  {
    String[] s = split(text, '\n');
    for (int i = 0; i < s.length; i++)
    {
      internalPrint(s[i], textColor);
      keepLine = false;
    }
  }

  /**
   * Starts a new line. Same as println(text) with empty text.
   */
  public synchronized void println()
  {
    println("");
  }

  /**
   * Appends given text to last line, if print() was previously called or
   * start a new line, if no print-method or println() was previously called.
   */
  public synchronized void print(String text)
  {
    String[] s = split(text, '\n');
    for (int i = 0; i < s.length; i++)
    {
      internalPrint(s[i], textColor);
      if (i != s.length - 1)
        keepLine = false;
      else
        keepLine = true;
    }
  }

  /**
   * Prints a boolean.
   */
  public synchronized void print(boolean b)
  {
    print(new Boolean(b).toString());
  }

  /**
   * Prints a character.
   */
  public synchronized void print(char ch)
  {
    print(new Character(ch).toString());
  }

  /**
   * Prints a array of characters.
   */
  public synchronized void print(char[] s)
  {
    print(new String(s));
  }

  /**
   * Prints a double-precision floating-point number.
   */
  public synchronized void print(double d)
  {
    print(new Double(d).toString());
  }

  /**
   * Prints a floating-point number.
   */
  public synchronized void print(float f)
  {
    print(new Float(f).toString());
  }

  /**
   * Prints an integer.
   */
  public synchronized void print(int i)
  {
    print(new Integer(i).toString());
  }

  /**
   * Prints a long integer.
   */
  public synchronized void print(long l)
  {
    print(new Long(l).toString());
  }

  /**
   * Prints a boolean and advances on next line.
   */
  public synchronized void println(boolean b)
  {
    println(new Boolean(b).toString());
  }

  /**
   * Prints a character and advances on next line.
   */
  public synchronized void println(char ch)
  {
    println(new Character(ch).toString());
  }

  /**
   * Prints an array of characters and advances on next line.
   */
  public synchronized void println(char[] s)
  {
    println(new String(s));
  }

  /**
   * Prints a double-precision floating-point number and advances on next line.
   */
  public synchronized void println(double d)
  {
    println(new Double(d).toString());
  }

  /**
   * Prints a floating-point number and advances on next line.
   */
  public synchronized void println(float f)
  {
    println(new Float(f).toString());
  }

  /**
   * Prints an integer and advances on next line.
   */
  public synchronized void println(int i)
  {
    println(new Integer(i).toString());
  }

  /**
   * Prints a long integer and advances on next line.
   */
  public synchronized void println(long l)
  {
    println(new Long(l).toString());
  }

  private void internalPrint(String text, int color)
  {
    if (keepLine)  // append
    {
      String s = (String)lines.lastElement();
      s += text;
      lines.setElementAt(s, lines.size()-1);
    }
    else // new line
    {
      lines.addElement(text);
      colors.addElement(new Integer(color));
    }

    if (nbLines > 0 && lines.size() > nbLines)
    {
      lines.removeElementAt(0);
      colors.removeElementAt(0);
    }

    startIndex = Math.max(0, lines.size() - displayLines);
    repaint();
  }

  /**
   * Erases all text and show empty console. Resets color of text to default.
   * The current color of the background is maintained.
   */
  public void clear()
  {
    clear(bgColor, textColor);
  }

  /**
   * Erases all text and show empty console and sets
   * background and text colors.
   */
  public void clear(int bgColor, int textColor)
  {
    lines.removeAllElements();
    colors.removeAllElements();
    int startIndex = 0;
    keepLine = false;
    this.textColor = textColor;
    this. bgColor = bgColor;
    repaint();
  }

  /**
   * Adds an OK soft button with label "OK". When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton()
  {
    addOkButton("OK");
  }

  /**
   * Adds an OK soft button with given label. When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton(String label)
  {
    okCommand = new Command(label, Command.OK, 1);
    addCommand(okCommand);
  }

  /**
   * Removes Exit soft button.
   * Gidlet's callback notification doExit() is disabled.
   */
  public void removeExitButton()
  {
   removeCommand(exitCommand);
  }

  /**
   * Removes Ok soft button.
   * Gidlet's callback notification doOk() is disabled.
   */
  public void removeOkButton()
  {
   removeCommand(okCommand);
  }

  // -------------------- Non public methods ---------------------------
  /**
   * Initialize MConsole. Called by main() if MConsole is created before main() is called.
   */
  protected void init()
  {
    addCommand(exitCommand);
    setCommandListener(new MyCommandListener());
    setFontSize(fontSize);

    font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, fSize);
    displayLines = getHeight() / font.getHeight();
    startIndex = 0;
    x = 0;
    action = -1;

    lines  = new Vector();
    colors = new Vector();

    keyThread = new KeyThread();
    keyThread.start();

    if (visible)
      show(true);
  }

  /**
   * Notificaton of key pressed.
   */
  protected void keyPressed(int key)
  {
    action = getGameAction(key);
    synchronized(keyThread)
    {
      keyThread.notify();
    }
  }

  /**
   * Notification of key repeated.
   */
  protected void keyRepeated(int key)
  {
    keyPressed(key);
  }

  /**
   * Notification of key release.
   */
  protected void keyReleased(int key)
  {
    action = -1;
  }

  /**
   * Used for rendering canvas.
   */
  protected void paint(Graphics g)
  {
    String string;
    Integer c;

    g.setFont(font);
    g.setColor(bgColor);
    g.fillRect(0, 0, getWidth(), getHeight());

    int count = Math.min(lines.size(), displayLines);
    int y = 0;

    for (int i = 0; i < count; i++)
    {
      string = (String)lines.elementAt(startIndex + i);
      c = (Integer)colors.elementAt(startIndex + i);

      g.setColor(c.intValue());
      g.drawString(string, x, y, ANCHOR);

      y += font.getHeight();
    }
  }

  private void setFontSize(int i)
  {
    switch (i)
    {
      case 1:
        fSize = Font.SIZE_SMALL;
        break;

      case 2:
        fSize = Font.SIZE_MEDIUM;
        break;

      case 3:
        fSize = Font.SIZE_LARGE;
        break;
    }
  }

  private String[] split(String s, char delimiter)
  {
    Vector v = new Vector();
    int startIndex = 0;
    int endIndex = 0;
    while (endIndex >= 0)
    {
      endIndex = s.indexOf(delimiter, startIndex);
      if (endIndex >= 0)
        v.addElement(s.substring(startIndex, endIndex));
      else
        v.addElement(s.substring(startIndex));
      startIndex = endIndex + 1;
    }
    String[] out = new String[v.size()];
    for (int i = 0; i < v.size(); i++)
      out[i] = (String)v.elementAt(i);
    return out;
  }
}
