// MForm.java

/*
  This software is part of Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
*/

package ch.aplu.gidlet;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.util.Hashtable;
import java.util.*;
import ch.aplu.gidlet.*;

/**
 * Form display with simple input and output items.
 *
 * A soft button 'Exit' is shown. Hitting the corresponding key will
 * call the Gidlet's doExit(). If you do not need this button,
 * invoke removeExitButton().<br>
 * If you implement your own CommandListener and call setCommandListener(), the internal
 * CommandListener is disabled.<br>
 *
 * main() serves as entry point for the user programm. It is called in the run-method
 * of a separate thread and should terminate to avoid non-terminated threads.<br>
 *
 * Creation may take place within main() or while declaring instance variable of
 * the application program.<br>
 *
 * If you register your own ItemStateListener, you will loose the internal one,
 * which is used to verify integer and double values entered with  addInputField().
 * If you want to maintain this feature you must call the
 * MForm.itemStateListener() in your own itemStateListener.<br>
 *
 * Example:<p>
 * <code><pre>
 *
 *import ch.aplu.gidlet.*;
 *import javax.microedition.lcdui.*;
 *
 *public class SimpleFormGidlet extends Gidlet
 *{
 *  private TextField tf1;
 *  private TextField tf2;
 *  private StringItem it;
 *
 *  private MForm form = new MForm("MForm example");
 *
 *  public void main()
 *  {
 *    tf1 = form.addInputField("Your first name:", "");
 *    tf2 = form.addInputField("Your family name:", "");
 *    it = form.addOutputField();
 *    form.addOkButton();
 *    form.show();
 *  }
 *
 *  public void doOk()
 *  {
 *    String val1 = form.getText(tf1);
 *    String val2 = form.getText(tf2);
 *    it.setText("Hello,\n" + val1 + " " + val2);
 *  }
 *}
 * </pre></code>
 */
public class MForm extends Form
{
// ------------------ Inner class MyCommandListener -------------
private class MyCommandListener implements CommandListener
{
  public void commandAction(Command c, Displayable d)
  {
    if (c == exitCommand)
      gidlet.doExit();
    if (okCommand != null && c == okCommand)
    {
      gidlet.doOk();
      gidlet.doOk(d);
      if (gidlet.isSleeping())
        gidlet.wakeUp();
    }
  }
}

// ------------------ Inner class MyItemStateListener -------------
private class MyItemStateListener implements ItemStateListener
{
  public void itemStateChanged(Item item)
  {
    if (item instanceof TextField)
    {
      TextField tf = (TextField)item;
      int validationFlag = ((Integer)validationTable.get(tf)).intValue();
      if (validationFlag == Gidlet.NONE_VALIDATE)
        return;

      String str = tf.getString();
      if (str.length() == 0)
        return;

      try
      {
        switch (validationFlag)
        {
          case Gidlet.INT_VALIDATE:
            Integer.parseInt(str);
            break;
          case Gidlet.LONG_VALIDATE:
            Long.parseLong(str);
            break;
          case Gidlet.DOUBLE_VALIDATE:
            Double.parseDouble(str);
            break;
          default:
            return;
        }
      }
      catch (NumberFormatException ex)
      {
        if (str.length() == 1 &&
            (str.charAt(0) == '+' || str.charAt(0) == '-'))
          return;
        str = str.substring(0, str.length()-1);
        tf.setString(str);
        AlertType.WARNING.playSound(gidlet.display);
      }
    }
  }
}

// ------------------ Inner class TimerTask ---------------------
private class MyTimerTask extends TimerTask
{
  public void run()
  {
    waitVisible = false;
  }
}
// ------------------ End of inner classes ----------------------

  private Gidlet gidlet;
  private Command exitCommand = new Command("Exit", Command.EXIT, 1);
  private Command okCommand = null;
  private Hashtable validationTable = new Hashtable();
  private boolean waitVisible;
  private MyItemStateListener isl;

  /**
   * Constructs a MForm with given title.
   * The display is initially invisible and should be made visible
   * by calling show().
   */
  public MForm(String title)
  {
    super(title);
    this.gidlet = Gidlet.myGidlet;
    isl = new MyItemStateListener();
    addCommand(exitCommand);
    setItemStateListener(isl);
    setCommandListener(new MyCommandListener());
  }

  /**
   * Constructs a MForm with no title.
   * The display is initially invisible and should be made visible
   * by calling show().
   */
  public MForm()
  {
    this("");
  }

  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * Because it is not garanteed that the diplay is actually visible when
   * Display.setCurrent() returns, wait for a maximum of 5 s for the display
   * to get shown.
   *
   */
  public void show()
  {
    show(true);
  }

  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * If wait = true, block until the display becomes actually visible, up
   * to a maximum auf 5 seconds. This may be useful to observe the start of
   * an animation.
   */
  public void show(boolean wait)
  {
    gidlet.display.setCurrent(this);
    if (wait)
    {
      waitVisible = true;
      Timer t = new Timer();
      t.schedule(new MyTimerTask(), 5000);
      while (waitVisible && !isShown())
        Gidlet.delay(100);
      t.cancel();
    }
  }

  /**
   * Adds a input text field to the MForm with given prompt, given initial value and given
   * validation type.
   * For validation types consult documentation for class ch.aplu.gidlet.Gidlet.
   * @see ch.aplu.gidlet.Gidlet
   * @return Reference to the created instance of TextField (subclass of Item)
   */
  public TextField addInputField(String prompt, String init, int validation)
  {
    int maxLen = 100;
    TextField in;
    switch (validation)
    {
      default:
      case Gidlet.NONE_VALIDATE:
        in = new TextField(prompt, init, maxLen, TextField.NON_PREDICTIVE);
        break;

      case Gidlet.INT_VALIDATE:
      case Gidlet.LONG_VALIDATE:
        in = new TextField(prompt, init, maxLen, TextField.NON_PREDICTIVE |
                                                 TextField.NUMERIC);
        break;

      case Gidlet.DOUBLE_VALIDATE:
        in = new TextField(prompt, init, maxLen, TextField.NON_PREDICTIVE |
                                                 TextField.DECIMAL);
        break;
    }
    validationTable.put(in, new Integer(validation));
    append(in);
    return in;
  }

  /**
   * Adds a input text field to the MForm with given prompt, given initial value
   * with no validation.
   * @return Reference to the created instance of TextField (subclass of Item)
   */
  public TextField addInputField(String prompt, String init)
  {
    return addInputField(prompt, init, Gidlet.NONE_VALIDATE);
  }

  /**
   * Adds a empty output text field to the MForm.
   * Text is written in the field with setText(). Nothing is seen until
   * some text is written.
   * If the text exceeds the height if the display, the
   * first line is on the top and the text may be scrolled using a small scroll
   * icon. Lines are separated by the plattform's line separator. If the length
   * of a line exceeds the width of the display, the line wraps at word bounds.<br>
   * Newline character '\n' is supported to force a line wrap.
   * @return Reference to the created instance of StringItem (subclass of Item)
   */
  public StringItem addOutputField()
  {
    Font font = Font.getFont(Font.FACE_SYSTEM,
                             Font.STYLE_PLAIN, Font.SIZE_MEDIUM);
    StringItem out = new StringItem(null, null);
    out.setFont(font);
    append(out);
    return out;
  }

  /**
   * Returns current text from the given TextField or StringItem,
   * null if not one of these.
   */
  public String getText(Item item)
  {
    if (item instanceof TextField)
      return ((TextField)item).getString();
    if (item instanceof StringItem)
      return ((StringItem)item).getText();
    return null;
  }

  /**
   * Writes given text to the given StringItem (subclass of Item).
   */
  public void setText(StringItem item, String text)
  {
    item.setText(text);
  }

  /**
   * Writes given text to the given Item (must be a StringItem).
   * (For backward compatibility only.)
   */
  public void setText(Item item, String text)
  {
    ((StringItem)item).setText(text);
  }

  /**
   * Calls the internal ItemStateListener  which is used to verify integer and
   * double values entered with  addInputField().
   * If you register your own ItemStateListener and you want to maintain the
   * verification, call this method in your own itemStateChanged().
   */
  public void itemStateChanged(Item item)
  {
    isl.itemStateChanged(item);
  }

  /**
   * Adds an OK soft button with label "OK". When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton()
  {
    addOkButton("OK");
  }

  /**
   * Adds an OK soft button with given label. When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton(String label)
  {
    okCommand = new Command(label, Command.OK, 1);
    addCommand(okCommand);
  }

  /**
   * Removes Exit soft button.
   * Gidlet's callback notification doExit() is disabled.
   */
  public void removeExitButton()
  {
   removeCommand(exitCommand);
  }

  /**
   * Removes Ok soft button.
   * Gidlet's callback notification doOk() is disabled.
   */
  public void removeOkButton()
  {
   removeCommand(okCommand);
  }
}
