// StringStore.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
*/

package ch.aplu.gidlet;

import java.io.*;
import javax.microedition.rms.*;


/**
 * Wrapper class for MIDlet's RMS (record managing system) based on class
 * RecordStore for storing strings exclusively. <br>
 * A record store belongs to the "MIDlet suite" identified by the MIDlet-name
 * and the MIDlet-vendor as defined in the application descriptor (JAD file).
 * It has a record store name that must be unique within the MIDlet suite.
 */
public class StringStore
{

// ----------------- Inner class StringRecord -----------------
private class StringRecord
{
  private int index;
  private String data;

  public StringRecord(int index, String data)
  {
    this.index = index;
    this.data = data;
  }

  public StringRecord(byte[] record)
  {
    ByteArrayInputStream raw = new ByteArrayInputStream(record);
    DataInputStream in = new DataInputStream(raw);
    try
    {
      index = in.readInt();
      data = in.readUTF();
    }
    catch (IOException ex) {}
  }

  public int getIndex()
  {
    return index;
  }

  public String getData()
  {
    return data;
  }

  public byte[] pack()
  {
    ByteArrayOutputStream raw = new ByteArrayOutputStream();
    DataOutputStream out = new DataOutputStream(raw);
    try
    {
      out.writeInt(index);
      out.writeUTF(data);
    }
    catch (IOException ex) {}
    return raw.toByteArray();
  }
}

// ----------------- Inner class MyRecordComparator -----------
private class MyRecordComparator implements RecordComparator
{
  public int compare(byte[] rec1, byte[] rec2)
  {
    int index1 = -1;
    int index2 = -1;
    try
    {
      ByteArrayInputStream raw1 = new ByteArrayInputStream(rec1);
      DataInputStream in1 = new DataInputStream(raw1);
      index1 = in1.readInt();
      ByteArrayInputStream raw2 = new ByteArrayInputStream(rec2);
      DataInputStream in2 = new DataInputStream(raw2);
      index2 = in2.readInt();
    }
    catch (IOException ex)
    {
      return -1;
    }
    if (index1 < index2)
      return RecordComparator.PRECEDES;
    if (index1 > index2)
      return RecordComparator.FOLLOWS;
    return RecordComparator.EQUIVALENT;
  }

}
// ----------------- End of inner classes ---------------------



  private String storeName;

  /**
   * Creates a StringStore instance with given record store name.
   * Record store names are case sensitive and may consist of any combination
   * of between one and 32 Unicode characters inclusive.<br><br>
   *
   * No attempt is made to open the record store.
   */
  public StringStore(String storeName)
  {
    this.storeName = storeName;
  }

  /**
   * Returns content of the StringStore.
   * @return content as string, null if the record store does not exist.
  */
  public String read()
  {
    StringBuffer sb = new StringBuffer();
    RecordEnumeration re = null;
    try
    {
      RecordStore rs = RecordStore.
        openRecordStore(storeName, false);
      re = rs.enumerateRecords(null, new MyRecordComparator(), false);
      while (re.hasNextElement())
      {
        byte[] record = re.nextRecord();
        sb.append(new StringRecord(record).getData());
      }
      rs.closeRecordStore();
    }
    catch (RecordStoreException ex)
    {
      if (re != null)
        re.destroy();
      return null;
    }

    re.destroy();
    return sb.toString();
  }


  /**
   * Appends string data to the end of the record store.
   * If the record store does not exist, it will be created.
   * The record store is closed when finishing.
   * @return true, if successful, false if operation fails
   */
  public boolean append(String s)
  {
    int nextIndex;
    try
    {
      RecordStore rs = RecordStore.
        openRecordStore(storeName, true, RecordStore.AUTHMODE_ANY, true);
      int nextRecordID = rs.getNextRecordID();
      if (nextRecordID == 1)  // Empty store, ID begins with 1
        nextIndex = 0;  // We start indexing with 0
      else
        nextIndex = new StringRecord(rs.getRecord(nextRecordID - 1)).getIndex() + 1;
      byte[] recordData = new StringRecord(nextIndex, s).pack();
      rs.addRecord(recordData, 0, recordData.length);
      rs.closeRecordStore();
    }
    catch (RecordStoreException ex)
    {
      return false;
    }
    return true;
  }


  /**
   * Writes string data the record store.
   * If the record store exists, all data is cleared.
   * If the record store does not exist, it will be created.
   * The record store is closed when finishing.
   * @return true, if successful, false if operation fails
   */
  public boolean write(String s)
  {
    delete();
    return append(s);
  }

  /**
   * Deletes record store. Only an owner MIDlet may perform this
   * operation (same MIDlet-name and vendor-name in the application descriptor).
   * @return true, if successful, false if operation fails
   */
  public boolean delete()
  {
    try
    {
      RecordStore.deleteRecordStore(storeName);
    }
    catch (RecordStoreException ex)
    {
      return false;
    }
    return true;
  }

  /**
   * Checks if the record store exists.
   * @return true, if record store exists, false otherwise.
   */
  public boolean exists()
  {
    try
    {
      RecordStore rs = RecordStore.
        openRecordStore(storeName, false);
    }
    catch (RecordStoreException ex)
    {
      return false;
    }
    return true;
  }

  /**
   * Returns amount of additional room available for record store to grow.
   * @return free space (in bytes), -1 if record store does not exist
   */
  public int getAvailableSpace()
  {
    int availableSpace;
    try
    {
      RecordStore rs = RecordStore.
        openRecordStore(storeName, true, RecordStore.AUTHMODE_ANY, true);
      availableSpace = rs.getSizeAvailable();

      rs.closeRecordStore();
    }
    catch (RecordStoreException ex)
    {
      return -1;
    }
    return availableSpace;
  }

  /**
   * Returns amount of space, that the record store occupies.
   * @return used space (in bytes), -1 if record store does not exist
   */
  public int getUsedSpace()
  {
    int usedSpace;
    try
    {
      RecordStore rs = RecordStore.
        openRecordStore(storeName, true, RecordStore.AUTHMODE_ANY, true);
      usedSpace = rs.getSize();

      rs.closeRecordStore();
    }
    catch (RecordStoreException ex)
    {
      return -1;
    }
    return usedSpace;
  }
}
